import os
import sys
import shutil
import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Mapping, Optional
import asyncio


from autograder.testcase_utils.abstract_testcase import TestCaseResult
from autograder.testcase_utils.abstract_testcase import TestCase as AbstractTestCase
from autograder.testcase_utils.shell import ShellCommand, EMPTY_COMMAND, ShellError, get_shell_command
from autograder.util import hide_path_to_directory

INCLUDE_MEMLEAK: str = '\n#include "leak_detector_c.h"\n'
MEMLEAK_DIR: Optional[TemporaryDirectory] = None
PRECOMPILED_MEMLEAK_FNAME = "memleak_detector.o"
MEMLEAK_SOURCE = Path(__file__).parent / "memleak" / "memleak_detector.c"
MEMLEAK_HEADER = Path(__file__).parent / "memleak" / "memleak_detector.h"
L = logging.getLogger("AUTOGRADER.ttypes.gcc")


class TestCase(AbstractTestCase):
    source_suffix = ".c"
    executable_suffix = ".out"
    helper_module = "test_helper.c"  # type: ignore
    compiler = get_shell_command("gcc")
    SUBMISSION_COMPILATION_ARGS: List[str] = ["-Dmain=__student_main__"]
    TESTCASE_COMPILATION_ARGS: List[str] = []
    MEMLEAK_TEMP_DIR: Optional[TemporaryDirectory] = None

    if sys.platform == "win32":
        # Windows cannot load its libraries if linking is dynamic
        TESTCASE_COMPILATION_ARGS += ["-static"]
    else:
        # Windows does not support re-declaring scanf_s even if mingw is used
        SUBMISSION_COMPILATION_ARGS += ["-Dscanf_s=scanf"]

    @classmethod
    def is_installed(cls) -> bool:
        return cls.compiler is not EMPTY_COMMAND

    @classmethod
    def get_template_dir(cls):
        return cls.type_source_file.parent / "c_templates"

    @classmethod
    async def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        cli_args: str,
        config: Mapping[str, Any],
        lock: asyncio.Lock,
    ) -> Path:
        """Compiles student submission without linking it to speed up total compilation time"""

        cli_args_lst = cli_args.split()
        if memleak_is_enabled(config):
            L.debug("Detected memleak to be enabled")
            async with lock:
                if not cls.MEMLEAK_TEMP_DIR:
                    await cls.precompile_memleak_detector(cls.TESTCASE_COMPILATION_ARGS)
                # Yeah, windows can't do anything in time
            if sys.platform == "win32":
                await asyncio.sleep(0.5)
            shutil.copy(Path(cls.MEMLEAK_TEMP_DIR.name) / PRECOMPILED_MEMLEAK_FNAME, student_dir)
            cli_args_lst.extend(["-include", str(MEMLEAK_HEADER)])

        copied_submission = await super().precompile_submission(
            submission,
            student_dir,
            [submission.stem],
            cli_args,
            config,
            lock,
        )
        precompiled_submission = copied_submission.with_suffix(".o")
        try:
            await cls.compiler(
                "-c",
                copied_submission,
                "-o",
                precompiled_submission,
                *cli_args_lst,
                *cls.SUBMISSION_COMPILATION_ARGS,
            )
        finally:
            copied_submission.unlink()
        return precompiled_submission

    async def precompile_testcase(self, cli_args: str):
        await self.compiler("-c", self.path, "-o", self.path.with_suffix(".o"), *cli_args.split())
        self.path.unlink()
        self.path = self.path.with_suffix(".o")

    async def compile_testcase(self, precompiled_submission: Path, cli_args: str) -> ShellCommand:
        executable_path = self.make_executable_path(precompiled_submission)
        files_to_compile = [
            precompiled_submission.with_name(self.path.name),
            precompiled_submission,
        ]
        if memleak_is_enabled(self.config):
            files_to_compile.append(precompiled_submission.with_name(PRECOMPILED_MEMLEAK_FNAME))
        await self.compiler(
            "-o",
            executable_path,
            *files_to_compile,
            *cli_args.split(),
            *self.TESTCASE_COMPILATION_ARGS,
        )

        return ShellCommand(executable_path)

    @classmethod
    async def precompile_memleak_detector(cls, compilation_args: List[Any]):
        cls.MEMLEAK_TEMP_DIR = TemporaryDirectory()
        tmp = Path(cls.MEMLEAK_TEMP_DIR.name)
        L.debug(f"CREATED TMP DIR FOR MEMLEAK, {cls.MEMLEAK_TEMP_DIR.name}")
        await cls.compiler("-c", MEMLEAK_SOURCE, "-o", tmp / PRECOMPILED_MEMLEAK_FNAME, *compilation_args)

    async def _weightless_run(
        self,
        precompiled_submission: Path,
        compiled_testcase: ShellCommand,
        testcase_runtime_args: str,
    ) -> TestCaseResult:
        if not memleak_is_enabled(self.config):
            return await super()._weightless_run(precompiled_submission, compiled_testcase, testcase_runtime_args)
        with TemporaryDirectory() as tmp:
            memleak_output = Path(tmp) / "memleak.txt"
            L.debug(f"Created memleak output file {memleak_output}")
            compiled_testcase = add_env_vars(compiled_testcase, MEMLEAK_FILE=str(memleak_output))
            result = await super()._weightless_run(precompiled_submission, compiled_testcase, testcase_runtime_args)
            memleak_output_text = memleak_output.read_text() if memleak_output.is_file() else ""
            memleak_output_text = hide_path_to_directory(memleak_output_text, precompiled_submission.parent)
        result.extra_output_fields["memleak_summary"] = memleak_output_text
        return result


def add_env_vars(command: ShellCommand, **env: str) -> ShellCommand:
    async def inner(*args, **kwargs):
        if not "env" in kwargs:
            kwargs["env"] = {}
        kwargs["env"].update(env)
        return await command(*args, **kwargs)

    return inner


def memleak_is_enabled(config: Mapping) -> bool:
    return config.get("GCC", {}).get("MEMORY_LEAK_DETECTION", False)



import os
import shutil
import stat
from pathlib import Path
from typing import Dict, Type

from .autograder import AutograderPaths
from .testcase_utils.abstract_testcase import TestCase
from .testcase_utils.testcase_picker import TestCasePicker


def create_dir(path: Path):
    if not path.exists():
        print(f"{path.name} directory does not exist. Creating...")
        path.mkdir()
    else:
        print(f"Found {path.name} directory")


def main(paths: AutograderPaths):
    if not paths.current_dir.exists() or not paths.current_dir.is_dir():
        print(f"Directory {paths.current_dir} not found. Please, use an existing directory.")
        exit(0)
    print("Hello. I will now guide you through the initial setup of autograder.")
    ans = input(f"Would you like to grade submissions located in '{paths.current_dir}'? (Yes/No) ")
    if not ans.lower().startswith("y"):
        print("You probably haven't specified a directory to the grader. Use 'autograder path/to/submissions/dir'")
        exit(0)
    create_dir(paths.tests_dir)
    create_dir(paths.testcases_dir)
    create_dir(paths.input_dir)
    create_dir(paths.output_dir)
    create_dir(paths.extra_dir)
    config_path = paths.config
    if not config_path.exists():
        print(f"config.ini not found in {paths.tests_dir}. Creating a default config...")
        paths.generate_config()
    else:
        print("Found config.ini")
    stdout_formatters_path = paths.stdout_formatters
    if not stdout_formatters_path.exists():
        print(f"{stdout_formatters_path.name} not found. Creating a default file...")
        shutil.copy(str(paths.default_stdout_formatters), str(paths.stdout_formatters))
    else:
        print("Found stdout_formatters.py")

    ans = input(
        "You are now ready to start working with autograder.\n"
        "Would you like me to generate the testcase templates? (Yes/No) "
    )
    if ans.lower().startswith("y"):
        supported_languages = _get_supported_languages(paths.testcase_types_dir)
        allowed_languages = ", ".join(name for name in supported_languages.keys())
        while True:
            choice = input(
                f"Choose a programming language you'd like to generate testcase templates for ({allowed_languages}): "
            )
            lang = supported_languages.get(choice, None)
            if lang is None:
                print(f"Couldn't find the language with name '{choice}'. Please, try again.")
            else:
                break
        safe_copytree(lang.get_template_dir(), paths.testcases_dir, dirs_exist_ok=True)
    print(
        "\n\nNow if you want to grade your submissions, you can use 'autograder run path/to/submissions/dir' "
        "for this directory."
    )
    print(f"You can write your testcases in {paths.testcases_dir}")
    print("If you want to see command line options, use 'autograder -h'")
    print(f"You can put the stdin inputs to your testcases into {paths.input_dir}")
    print(f"You can put the expected outputs to your testcases into {paths.output_dir}")
    print(f"You can put the extra files to be available for each testcase into {paths.extra_dir}")
    print(f"You can configure grading by editing {paths.config}")
    print("You can find readme at https://github.com/Ovsyanka83/autograder")


def _get_supported_languages(testcase_types_dir: Path) -> Dict[str, Type[TestCase]]:
    testcase_types = TestCasePicker.discover_testcase_types(testcase_types_dir)
    return {t.type_source_file.stem: t for t in testcase_types if (t.get_template_dir()).exists()}


def safe_copytree(
    src,
    dst,
    symlinks=False,
    ignore=None,
    copy_function=shutil.copy2,
    ignore_dangling_symlinks=False,
    dirs_exist_ok=False,
):
    """This is a newer version of copytree only available for python3.8
    We need it because it has a dirs_exist_ok keyword arg
    """
    with os.scandir(src) as itr:
        entries = list(itr)
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    os.makedirs(dst, exist_ok=dirs_exist_ok)
    errors = []
    use_srcentry = copy_function is shutil.copy2 or copy_function is shutil.copy

    for srcentry in entries:
        if srcentry.name in ignored_names:
            continue
        src_name = os.path.join(src, srcentry.name)
        dst_name = os.path.join(dst, srcentry.name)
        src_obj = srcentry if use_srcentry else src_name
        try:
            is_symlink = srcentry.is_symlink()
            if is_symlink and os.name == "nt":
                # Special check for directory junctions, which appear as
                # symlinks but we want to recurse.
                lstat = srcentry.stat(follow_symlinks=False)
                if lstat.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT:
                    is_symlink = False
            if is_symlink:
                linkto = os.readlink(src_name)
                if symlinks:
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(linkto, dst_name)
                    shutil.copystat(src_obj, dst_name, follow_symlinks=not symlinks)
                else:
                    # ignore dangling symlink if the flag is on
                    if not os.path.exists(linkto) and ignore_dangling_symlinks:
                        continue
                    # otherwise let the copy occur. copy2 will raise an error
                    if srcentry.is_dir():
                        shutil.copytree(
                            src_obj,
                            dst_name,
                            symlinks,
                            ignore,
                            copy_function,
                            dirs_exist_ok=dirs_exist_ok,  # type: ignore
                        )
                    else:
                        copy_function(src_obj, dst_name)
            elif srcentry.is_dir():
                shutil.copytree(src_obj, dst_name, symlinks, ignore, copy_function, dirs_exist_ok=dirs_exist_ok)  # type: ignore
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(src_obj, dst_name)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((src_name, dst_name, str(why)))
    try:
        shutil.copystat(src, dst)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, "winerror", None) is None:
            errors.append((src, dst, str(why)))
    if errors:
        raise shutil.Error(errors)
    return dst



import asyncio
import itertools
import shutil
import sys
import time
from pathlib import Path
from tempfile import TemporaryDirectory
import logging
from typing import Callable, Dict, List, Optional, Set, Tuple, Type, Union

from .config_manager import GradingConfig
from .output_summary import GradingOutputLogger, JsonGradingOutputLogger
from .testcase_utils.abstract_testcase import TestCase
from .testcase_utils.shell import ShellError
from .testcase_utils.stdout_testcase import StdoutOnlyTestCase
from .testcase_utils.submission import Submission, find_appropriate_source_file_stem
from .testcase_utils.testcase_io import TestCaseIO
from .testcase_utils.testcase_picker import TestCasePicker
from .util import AutograderError, get_file_names, hide_path_to_directory, import_from_path

EMPTY_TESTCASE_IO = TestCaseIO.get_empty_io()


L = logging.getLogger("AUTOGRADER.grader")


# TODO: What if grader built a less complex object whose only purpose is actually grading? Should improve complexity.
class Grader:
    json_output: bool
    submissions: List[Submission]
    tests: Dict[Type[TestCase], List[TestCase]]
    raw_submissions: List[str]
    paths: "AutograderPaths"
    temp_dir: Path
    _temp_dir: TemporaryDirectory

    logger: GradingOutputLogger
    config: GradingConfig
    testcase_picker: TestCasePicker

    def __init__(self, current_dir: Path, json_output: bool = False, submissions: Optional[List[str]] = None):
        if submissions is None:
            submissions = []
        self.json_output = json_output
        self.raw_submissions = submissions
        self.stdout_formatters = {}
        self.paths = AutograderPaths(current_dir)
        self.config = GradingConfig(self.paths.config, self.paths.default_config)
        self.testcase_picker = TestCasePicker(self.paths.testcase_types_dir, self.config.stdout_only_grading_enabled)
        self.stdout_formatters = self._import_formatters(self.paths.stdout_formatters)
        logger_type = JsonGradingOutputLogger if self.json_output else GradingOutputLogger
        self.logger = logger_type(
            self.paths.results_dir,
            self.config.assignment_name,
            self.config.total_points_possible,
            self.config.generate_results,
        )

    def run(self):
        io_choices = {}
        try:
            self._prepare_directory_structure()
            self.submissions = self._gather_submissions()
            io_choices = self._gather_io()
            # also copies testcase_utils to temp dir
            self.tests, unused_io_choices = self._gather_testcases(io_choices.copy())
            if self.config.stdout_only_grading_enabled:
                self.tests[StdoutOnlyTestCase] = self._generate_stdout_only_testcases(unused_io_choices)

            # Allows for consistent output
            for test_list in self.tests.values():
                test_list.sort(key=lambda t: t.path.name)

            precompilation_tasks = itertools.chain.from_iterable(
                (t.precompile_testcase(self.config.testcase_precompilation_args[t.name]) for t in t_lst)
                for t_lst in self.tests.values()
            )
            tasks = map(Runner(self, asyncio.Lock()), self.submissions)

            if sys.platform == "win32":
                loop = asyncio.ProactorEventLoop()
                asyncio.set_event_loop(loop)
            asyncio.get_event_loop().run_until_complete(asyncio.gather(*precompilation_tasks))
            modified_submissions = asyncio.get_event_loop().run_until_complete(asyncio.gather(*tasks))
            total_class_points = sum(s.final_grade for s in modified_submissions)
            class_average = round(total_class_points / len(self.submissions))
            self.logger.print_final_score(modified_submissions, class_average)
            self.logger.print_key()
        finally:
            for io in io_choices.values():
                io.cleanup()
            self.cleanup()
        return class_average

    def cleanup(self):
        if self.temp_dir.exists():
            # Windows doesn't know how to clean processes in time...
            if sys.platform == "win32":
                time.sleep(1)
            self._temp_dir.cleanup()

    def _prepare_directory_structure(self):
        self._temp_dir = TemporaryDirectory()
        self.temp_dir = Path(self._temp_dir.name)
        self._check_required_directories_exist()
        if self.config.generate_results:
            self.paths.results_dir.mkdir(exist_ok=True)

    def _check_required_directories_exist(self):
        for directory in self.paths.required_dirs:
            if not directory.exists():
                raise AutograderError(
                    f"{directory} directory not found. It is required for the grader to function.\n"
                    "Maybe you specified the incorrect directory? Use `autograder submission_directory_here`"
                )

    def _gather_submissions(self) -> List[Submission]:
        """Returns sorted list of paths to submissions"""
        submissions_to_grade = set(self.raw_submissions)
        submissions: List[Submission] = []
        for submission_path in self.paths.current_dir.iterdir():
            if submissions_to_grade and submission_path.name not in submissions_to_grade:
                continue

            testcase_type = self.testcase_picker.pick(submission_path, self.config.possible_source_file_stems)
            if testcase_type is not None:
                if (
                    self.config.any_submission_file_name_is_allowed
                    or find_appropriate_source_file_stem(submission_path, self.config.possible_source_file_stems)
                    is not None
                ):
                    submissions.append(Submission(submission_path, testcase_type, self.temp_dir))
                else:
                    pass

        if not len(submissions):
            raise AutograderError(f"No student submissions found in '{self.paths.current_dir}'.")

        # Allows consistent output
        submissions.sort(key=lambda s: s.old_path)
        return submissions

    def _gather_io(self) -> Dict[str, TestCaseIO]:
        outputs = get_file_names(self.paths.output_dir)
        inputs = get_file_names(self.paths.input_dir)
        io: Set[str] = set().union(outputs, inputs)
        io_set = {Path(i) for i in io}
        dict_io = {
            p.stem: TestCaseIO(
                p,
                self.stdout_formatters,
                self.paths.input_dir,
                self.paths.output_dir,
            )
            for p in io_set
        }
        return dict_io

    def _generate_stdout_only_testcases(self, io_choices: Dict[str, TestCaseIO]) -> List[TestCase]:
        return [
            StdoutOnlyTestCase(
                io.output_file,
                self.config.timeouts[io.name],
                self.config.testcase_weights[io.name],
                io,
                self.config.file,
            )
            for io in io_choices.values()
            if io.expected_output
        ]

    def _gather_testcases(
        self,
        io: Dict[str, TestCaseIO],
    ) -> Tuple[Dict[Type[TestCase], List[TestCase]], Dict[str, TestCaseIO]]:
        """Returns sorted list of testcase_types from tests/testcases"""
        if not self.paths.testcases_dir.exists():
            return {}, io
        tests: Dict[Type[TestCase], List[TestCase]]
        tests = {t: [] for t in self.testcase_picker.testcase_types}
        for test in self.paths.testcases_dir.iterdir():
            if not test.is_file():
                continue
            testcase_type = self.testcase_picker.pick(test, self.config.possible_source_file_stems)
            if testcase_type is None:
                continue
            shutil.copy(str(test), str(self.temp_dir))
            tests[testcase_type].append(
                testcase_type(
                    self.temp_dir / test.name,
                    self.config.timeouts[test.name],
                    self.config.testcase_weights[test.name],
                    io.pop(test.stem, EMPTY_TESTCASE_IO),
                    self.config.file,
                )
            )
        return tests, io

    @staticmethod
    def _import_formatters(path_to_stdout_formatters: Path) -> Dict[str, Callable[[str], str]]:
        if path_to_stdout_formatters.exists():
            module = import_from_path("stdout_formatters", path_to_stdout_formatters)
            return {k: v for k, v in module.__dict__.items() if callable(v)}
        else:
            return {}


class AutograderPaths:
    __slots__ = (
        "current_dir",
        "results_dir",
        "output_summary",
        "tests_dir",
        "testcases_dir",
        "extra_dir",
        "input_dir",
        "output_dir",
        "stdout_formatters",
        "testcase_types_dir",
        "default_stdout_formatters",
        "config",
        "default_config",
        "required_dirs",
    )
    current_dir: Path
    results_dir: Path
    output_summary: Path
    tests_dir: Path

    testcases_dir: Path
    extra_dir: Path
    input_dir: Path
    output_dir: Path
    stdout_formatters: Path
    config: Path
    required_dirs: Tuple[Path, ...]

    testcase_types_dir: Path
    default_stdout_formatters: Path
    default_config: Path

    def __init__(self, current_dir: Union[Path, str]):
        current_dir = Path(current_dir)
        self.current_dir = current_dir
        self.results_dir = current_dir / "results"
        self.output_summary = current_dir / "grader_output.txt"
        self.tests_dir = current_dir / "tests"

        self.testcases_dir = self.tests_dir / "testcases"
        self.extra_dir = self.tests_dir / "extra"
        self.input_dir = self.tests_dir / "input"
        self.output_dir = self.tests_dir / "output"

        self.stdout_formatters = self.tests_dir / "stdout_formatters.py"
        self.config = self.tests_dir / "config.toml"

        self.required_dirs = ()

        autograder_dir = Path(__file__).parent
        self.testcase_types_dir = autograder_dir / "testcase_types"
        self.default_stdout_formatters = autograder_dir / "default_stdout_formatters.py"
        self.default_config = autograder_dir / "default_config.toml"

    def generate_config(self):
        if not self.config.exists():
            if self.default_config.exists():
                shutil.copy(str(self.default_config), str(self.config))
            else:
                raise AutograderError(f"Failed to generate config:'{self.default_config}' not found.")


class Runner:
    """Grades single submissions"""

    grader: Grader
    lock: asyncio.Lock

    def __init__(self, grader: Grader, lock: asyncio.Lock):
        self.grader = grader
        self.lock = lock

    async def __call__(self, submission: Submission) -> Submission:
        await self.run_on_single_submission(submission, self.lock)
        return submission

    async def run_on_single_submission(self, submission: Submission, lock: asyncio.Lock) -> None:
        self._copy_extra_files(submission.temp_dir)
        await self._get_testcase_output(submission)
        async with lock:
            self.grader.logger.print_single_student_grading_results(submission)
        # Windows sucks at cleaning up processes early
        if not sys.platform.startswith("win32"):
            # Cleanup after running tests on student submission
            shutil.rmtree(submission.temp_dir)

    def _copy_extra_files(self, to_dir: Path) -> None:
        if self.grader.paths.extra_dir.exists():
            for path in self.grader.paths.extra_dir.iterdir():
                new_path = to_dir / path.name
                shutil.copy(str(path), str(new_path))

    # TODO: rename to "_run_testcases_on_submission" or something
    async def _get_testcase_output(self, submission: Submission):
        """Grades single submission and returns its normalized score

        Note: Has side effects in submission instance
        """
        precompilation_lock = asyncio.Lock()
        try:
            precompiled_submission = await submission.type.precompile_submission(
                submission.old_path,
                submission.temp_dir,
                self.grader.config.possible_source_file_stems,
                self.grader.config.submission_precompilation_args[submission.old_path.name],
                self.grader.config.file,
                precompilation_lock,
            )
        except ShellError as e:
            error = hide_path_to_directory(e.format("Failed to precompile:"), submission.temp_dir)
            submission.register_precompilation_error(error)
            return 0
        allowed_tests = self.grader.tests.get(submission.type, [])
        if not allowed_tests:
            submission.register_precompilation_error("No suitable testcases found.")
            return 0
        submission.type.run_additional_testcase_operations_in_student_dir(submission.temp_dir)
        for test in allowed_tests:
            result = await test.run(
                precompiled_submission,
                self.grader.config.testcase_compilation_args[test.name],
                self.grader.config.testcase_runtime_args[test.name],
            )
            message = hide_path_to_directory(result.message, submission.temp_dir)
            submission.add_grade(test.name, result.grade, test.weight, message, result.extra_output_fields)
        submission.register_final_grade(self.grader.config.total_score_to_100_ratio)



import re
from pathlib import Path
from typing import Any, Dict

from .exit_codes import ExitCodeEventType
from .testcase_result_validator import LAST_LINE_SPLITTING_CHARACTER

template_matcher = re.compile("{ *% *([A-Za-z0-9_]+) *% *}")

FORMAT_KWARGS = {
    "RESULT_EXIT_CODE": str(int(ExitCodeEventType.RESULT)),
    "CHECK_STDOUT_EXIT_CODE": str(int(ExitCodeEventType.CHECK_STDOUT)),
    "CHEAT_ATTEMPT_EXIT_CODE": str(int(ExitCodeEventType.CHEAT_ATTEMPT)),
    "SPLITCHAR": LAST_LINE_SPLITTING_CHARACTER,
}


def format_template(template: str, safe: bool = True, **kwargs: Dict[str, Any]):
    # We use dict here to filter repeated matches
    matches = {m.group(1): m.group(0) for m in template_matcher.finditer(template)}
    for attr, matched_string in matches.items():
        value = kwargs.pop(attr, None)
        if value is None:
            raise ValueError(f"Attribute '{attr}' not supplied")
        template = template.replace(matched_string, str(value))
    if len(kwargs) and safe:
        raise ValueError("Too many arguments supplied: " + ", ".join(kwargs.keys()))
    return template


def get_formatted_test_helper(path_to_test_helper: Path, **exta_format_kwargs: str) -> str:
    with path_to_test_helper.open() as helper_file:
        return format_template(helper_file.read(), safe=False, **FORMAT_KWARGS, **exta_format_kwargs)



from pathlib import Path
from typing import Callable, Dict

from ..config_manager import DEFAULT_ARGLIST_VALUE_KEY

FORMATTER_TYPE = Callable[[str], str]


class TestCaseIO:
    """Represents everything related to stdin/stdout in testcase_utils"""

    stem: str
    formatter: FORMATTER_TYPE

    def __init__(
        self,
        testcase_path: Path,  # A path of either input or output
        formatters: Dict[str, FORMATTER_TYPE],
        input_dir: Path,
        output_dir: Path,
    ):
        self.name = testcase_path.name
        self.stem = testcase_path.stem
        f = formatters  # to make next line shorter
        self.formatter = f.get(self.stem) or f.get(DEFAULT_ARGLIST_VALUE_KEY) or default_formatter
        self.output_file: Path = output_dir / f"{self.stem}.txt"

        if self.output_file.exists() and self.output_file.is_file():
            self.expected_output_backup = self.output_file.read_text()
            self.expected_output = self.format_output(self.expected_output_backup)
            self.output_file.write_text("")
        else:
            self.expected_output_backup = ""
            self.expected_output = ""

        input_file: Path = input_dir / f"{self.stem}.txt"
        if input_file.exists() and input_file.is_file():
            self.input = input_file.read_text().strip()
        else:
            self.input = ""

    def format_output(self, output: str) -> str:
        # Replacement of \r\n accounts for inconsistent line endings between systems
        return self.formatter(output).replace("\r\n", "\n")

    def expected_output_equals(self, output: str) -> bool:
        return self.format_output(output) == self.expected_output

    @classmethod
    def get_empty_io(cls):
        io = cls(Path("nonexistent"), {}, Path(), Path())
        io.expected_output = ""
        io.input = ""
        return io

    def cleanup(self):
        if self.output_file.exists() and self.output_file.is_file():
            self.output_file.write_text(self.expected_output_backup)


def default_formatter(s: str) -> str:
    return s



from autograder.testcase_types.gcc.c import TestCase as CTestCase
from autograder.testcase_utils.shell import get_shell_command


class TestCase(CTestCase):
    source_suffix = ".cpp"
    compiler = get_shell_command("g++")

    @classmethod
    def get_template_dir(cls):
        return cls.type_source_file.parent / "c++_templates"



from pathlib import Path
from typing import List, Optional, Type

from autograder.util import AutograderError, import_from_path

from .abstract_testcase import TestCase
from .stdout_testcase import StdoutOnlyTestCase


class TestCasePicker:
    testcase_types: List[Type[TestCase]]

    def __init__(self, testcase_types_dir: Path, stdout_only_grading_enabled: bool = False):
        self.testcase_types = self.discover_testcase_types(testcase_types_dir)
        if stdout_only_grading_enabled:
            self.testcase_types.insert(0, StdoutOnlyTestCase)
        if not self.testcase_types:
            raise AutograderError("No acceptable testcase types were detected.\n")

    @classmethod
    def discover_testcase_types(cls, testcase_types_dir: Path) -> List[Type[TestCase]]:
        testcase_types: List[Type[TestCase]] = []
        for testcase_type_dir in testcase_types_dir.iterdir():
            for path in testcase_type_dir.iterdir():
                if path.is_file() and path.suffix == ".py":
                    module = import_from_path(f"testcase:{path.stem}{testcase_type_dir.name}", path)
                    testcase_type: Optional[Type[TestCase]] = getattr(module, "TestCase", None)
                    if testcase_type is None:
                        continue

                    if testcase_type.is_installed():
                        testcase_types.append(testcase_type)
        return testcase_types

    def pick(self, file: Path, possible_source_file_stems: List[str]) -> Optional[Type[TestCase]]:
        for testcase_type in self.testcase_types:
            if testcase_type.is_a_type_of(file, possible_source_file_stems):
                return testcase_type



import shutil
import stat
import sys
from pathlib import Path
from typing import Any, Callable, List


from ..config_manager import GradingConfig
from .abstract_testcase import TestCase
from .exit_codes import ExitCodeEventType
from .shell import ShellCommand, ShellError, get_shell_command
from .submission import find_appropriate_source_file_stem
from .testcase_result_validator import LAST_LINE_SPLITTING_CHARACTER

POSSIBLE_MAKEFILE_NAMES = "GNUmakefile", "makefile", "Makefile"


def is_multifile_submission(submission_dir: Path, possible_source_file_stems: List[str]) -> bool:
    if not submission_dir.is_dir():
        return False
    contents = list(submission_dir.iterdir())
    # Seek the innermost non-single-file directory
    # in case the directory contains another directory with the project (common issue when zipping directories)
    while len(contents) == 1 and contents[0].is_dir():
        contents = list(contents[0].iterdir())

    for f in submission_dir.iterdir():
        if _is_makefile(f) or find_appropriate_source_file_stem(f, possible_source_file_stems) is not None:
            return True
    return False


def _is_makefile(path: Path):
    return not path.is_dir() and path.name in POSSIBLE_MAKEFILE_NAMES


def contains_shebang(path: Path) -> bool:
    if not path.is_file():
        return False
    # Windows does not support shebang lines
    if sys.platform.startswith("win32"):
        return False
    with open(path) as f:
        return f.readline().startswith("#!")


class StdoutOnlyTestCase(TestCase):

    helper_module = ""  # type: ignore
    compiler = get_shell_command("make")

    @classmethod
    def is_installed(cls) -> bool:
        """We assume all utilities necessary are installed because we can't possibly check for all of them."""
        return cls.compiler is not None

    @classmethod
    def is_a_type_of(cls, file: Path, possible_source_file_stems: List[str]) -> bool:
        return contains_shebang(file) or is_multifile_submission(file, possible_source_file_stems)

    @classmethod
    async def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        cli_args: str,
        config: GradingConfig,
    ):
        """pwd is temp/student_dir"""

        # Single-file with shebang
        if submission.is_file():
            destination = student_dir / submission.name
            shutil.copy(str(submission), str(destination))
            _make_executable(destination)
            return destination
        # A directory with a makefile
        else:
            _copy_multifile_submission_contents_into_student_dir(submission, student_dir)
            try:
                await cls.compiler(*cli_args.split(), cwd=student_dir)
            except ShellError as e:
                if "no makefile found" not in str(e):
                    raise e
            executable = _find_submission_executable(student_dir, possible_source_file_stems)
            if executable is None:
                raise ShellError(
                    -1,
                    "Submission was successfully precompiled but the executable could not be found. Most likely it was not in POSSIBLE_SOURCE_FILE_STEMS.",
                )
            return executable

    async def compile_testcase(self, precompiled_submission: Path, cli_args: str):
        return _add_args(self._run_stdout_only_testcase, precompiled_submission)

    def prepend_test_helper(self):
        """We don't need TestHelper when we are only checking inputs/outputs"""

    def delete_executable_files(self, precompiled_submission: Path):
        """We don't compile each testcase separately so we don't need to clean up executable files"""

    def delete_source_file(self, source_path: Path):
        """There is no testcase file"""

    async def _run_stdout_only_testcase(self, precompiled_submission: Path, *args: Any, **kwargs: Any):
        # Because student submissions do not play by our ExitCodeEventType rules,
        # we allow them to return 0 at the end.
        kwargs["allowed_exit_codes"] = (0,)
        result = await ShellCommand(precompiled_submission)(*args, **kwargs)

        # We fake the validation string because there is no way we can truly validate such testcases
        result.stdout += f"\n-1{LAST_LINE_SPLITTING_CHARACTER}{self.validating_string}"
        result.returncode = ExitCodeEventType.CHECK_STDOUT
        return result


def _copy_multifile_submission_contents_into_student_dir(submission: Path, student_dir: Path):
    contents: List[Path] = list(submission.iterdir())
    while len(contents) == 1 and contents[0].is_dir():
        contents = list(contents[0].iterdir())
    for f in contents:
        new_path = student_dir / f.name
        if f.is_dir():
            op = shutil.copytree
        else:
            op = shutil.copyfile
        op(str(f), new_path)


def _find_submission_executable(student_dir: Path, possible_source_file_stems: List[str]):
    for f in student_dir.iterdir():
        if find_appropriate_source_file_stem(f, possible_source_file_stems):
            return f


def _make_executable(f: Path) -> None:
    f.chmod(f.stat().st_mode | stat.S_IEXEC)


def _add_args(function: Callable[..., Any], *initial_args: Any) -> Callable[..., Any]:
    async def inner(*args: Any, **kwargs: Any):
        return await function(*initial_args, *args, **kwargs)

    return inner



import asyncio
import dataclasses
import os
import shutil
import subprocess as synchronous_subprocess
import sys
import logging
from asyncio import subprocess
from concurrent.futures import TimeoutError
from dataclasses import dataclass
from locale import getpreferredencoding
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Union


L = logging.getLogger("AUTOGRADER.testcase_utils.shell")


@dataclass
class ShellCommandResult:
    returncode: int
    stdout: str
    stderr: str
    # This is a small hack to
    # extra_streams: Dict[str, str] = dataclasses.field(default_factory=dict)


@dataclass
class ShellCommand:
    """We use this class to abstract away from systemcall libraries"""

    command_name: Union[str, Path]

    async def __call__(
        self,
        *args: object,
        allowed_exit_codes: Sequence[int] = (0,),
        timeout: Optional[float] = None,
        stdin: str = "",
        **kwargs: Any,
    ) -> ShellCommandResult:
        os_specific_kwargs = {}
        if sys.platform == "win32":
            os_specific_kwargs["startupinfo"] = synchronous_subprocess.STARTUPINFO(
                dwFlags=synchronous_subprocess.STARTF_USESHOWWINDOW,
                wShowWindow=synchronous_subprocess.SW_HIDE,
            )
            if "env" in kwargs:
                kwargs["env"].update({"SYSTEMROOT": os.environ["SYSTEMROOT"]})
        process = await subprocess.create_subprocess_exec(
            # Linux handles non-string args well yet Windows doesn't
            str(self.command_name),
            *[str(a) for a in args],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            **kwargs,
            **os_specific_kwargs,
        )
        # That's the same way subprocess.Popen(text=True) gets the encoding
        encoding = getpreferredencoding(False)
        try:
            result = await asyncio.wait_for(process.communicate(input=stdin.encode(encoding)), timeout=timeout)
        except TimeoutError as e:
            # Windows doesn't know how to clean up its processes
            if process.returncode is None and sys.platform == "win32":
                # We could probably do this asynchronously but I am too lazy to test it
                synchronous_subprocess.run(
                    ["taskkill", "/F", "/T", "/PID", str(process.pid)],
                    **os_specific_kwargs,
                    stdout=synchronous_subprocess.DEVNULL,
                    stderr=synchronous_subprocess.DEVNULL,
                )
            raise e
        stdout, stderr = (s.decode(encoding) for s in result)
        L.debug(
            f"""({process.returncode}) EXECUTED CMD: {self.command_name} {' '.join([str(a) for a in args])}
                STDOUT: {stdout}
                STDERR: {stderr}
            """
        )
        returncode = process.returncode if process.returncode is not None else -1

        if process.returncode not in allowed_exit_codes:
            raise ShellError(returncode, stderr)
        return ShellCommandResult(returncode, stdout, stderr)


EMPTY_COMMAND = ShellCommand("false")


class ShellError(Exception):
    def __init__(self, returncode: int, stderr: str):
        super().__init__(stderr)
        self.returncode = returncode
        self.stderr = stderr

    def format(self, title: str) -> str:
        return f"{title}\n{self.stderr}"


def get_shell_command(command: str) -> ShellCommand:
    """An API for commands that postpone throwing non-existence errors from creation to runtime"""
    return EMPTY_COMMAND if shutil.which(command) is None else ShellCommand(command)



from dataclasses import dataclass
import re
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type

from .abstract_testcase import TestCase


def find_appropriate_source_file_stem(file: Path, possible_source_file_stems: List[str]) -> Optional[str]:
    for s in possible_source_file_stems:
        if s.lower() in file.stem.lower():
            return s


STUDENT_NAME_MATCHER = re.compile(r"(?P<student_name>[A-Za-z]+)_\d+_\d+_.+")


def get_submission_name(submission: Path):
    match = STUDENT_NAME_MATCHER.match(submission.stem)
    return submission.name if match is None else match["student_name"]


@dataclass
class TestCaseGrade:
    testcase_score: float
    testcase_weight: float
    message: str
    extra_output_fields: Dict[str, str]


class Submission:
    # Sowwy, dataclasses do not support slots until 3.10
    __slots__ = "old_path", "temp_path", "temp_dir", "name", "type", "grades", "precompilation_error", "final_grade"

    old_path: Path
    temp_path: Path
    temp_dir: Path  # Personal temporary dir, not the root one
    name: str
    type: Type[TestCase]
    grades: Dict[str, TestCaseGrade]
    precompilation_error: str
    final_grade: int

    def __init__(self, file: Path, testcase_type: Type[TestCase], temp_dir: Path):
        self.old_path = file
        self.temp_dir = temp_dir / file.name
        self.temp_dir.mkdir()
        self.temp_path = self.temp_dir / file.name
        self.name = get_submission_name(file)
        self.type = testcase_type
        self.grades = {}
        self.precompilation_error = ""
        self.final_grade = -1

    def add_grade(
        self,
        test_name: str,
        testcase_score: float,
        testcase_weight: float,
        message: str,
        extra_output_fields: Dict[str, str],
    ):
        self.grades[test_name] = TestCaseGrade(testcase_score, testcase_weight, message, extra_output_fields)

    def register_precompilation_error(self, error: str) -> None:
        self.precompilation_error = error
        self.final_grade = 0

    def register_final_grade(self, total_score_to_100_ratio: float):
        self.final_grade = self._calculate_final_grade(total_score_to_100_ratio)

    def _calculate_final_grade(self, total_score_to_100_ratio: float) -> int:
        total_score = 0
        total_testcase_weight = 0
        for grade in self.grades.values():
            total_score += grade.testcase_score
            total_testcase_weight += grade.testcase_weight
        normalized_score = total_score / total_testcase_weight * total_score_to_100_ratio
        return round(normalized_score)



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, List, Mapping, Optional, TypeVar

from tomlkit.api import parse
from tomlkit.container import Container

DEFAULT_FILE_STEM = "Homework"
MAIN_CONFIG_SECTION = "CONFIG"
DEFAULT_ARGLIST_VALUE_KEY = "DEFAULT"


TESTNAME = TypeVar("TESTNAME", bound=str)
VT = TypeVar("VT")


@dataclass
class ArgList(Generic[TESTNAME, VT]):
    mapping: Mapping[TESTNAME, VT]
    fallback_value: VT
    default_value_key: TESTNAME = DEFAULT_ARGLIST_VALUE_KEY  # type: ignore

    def __getitem__(self, key: TESTNAME) -> VT:
        if key in self.mapping:
            return self.mapping[key]
        elif self.default_value_key in self.mapping:
            return self.mapping[self.default_value_key]
        else:
            return self.fallback_value


class GradingConfig:
    file: Mapping[str, Any]

    timeouts: ArgList[str, float]
    generate_results: bool
    stdout_only_grading_enabled: bool
    total_points_possible: int
    total_score_to_100_ratio: float

    assignment_name: str
    any_submission_file_name_is_allowed: bool
    possible_source_file_stems: List[str]
    testcase_weights: ArgList[str, float]

    def __init__(self, config: Path, default_config: Path):
        self._configure_grading(config, default_config)

    def _configure_grading(
        self,
        config_path: Path,
        default_config_path: Path,
    ):
        global_config = _read_config(config_path, default_config_path)
        self.file = global_config
        cfg = global_config[MAIN_CONFIG_SECTION]
        self.timeouts = ArgList(cfg["TIMEOUT"], 1)
        self.generate_results = cfg["GENERATE_RESULTS"]
        self.stdout_only_grading_enabled = cfg["STDOUT_ONLY_GRADING_ENABLED"]

        self.total_points_possible = cfg["TOTAL_POINTS_POSSIBLE"]
        self.total_score_to_100_ratio = self.total_points_possible / 100

        self.assignment_name = cfg["ASSIGNMENT_NAME"]

        self.possible_source_file_stems = cfg["POSSIBLE_SOURCE_FILE_STEMS"]
        if not self.possible_source_file_stems:
            self.possible_source_file_stems = [DEFAULT_FILE_STEM]
            self.any_submission_file_name_is_allowed = True
        else:
            self.any_submission_file_name_is_allowed = False

        self.testcase_weights = ArgList(cfg["TESTCASE_WEIGHT"], 1)

        self.submission_precompilation_args = ArgList(cfg["SUBMISSION_PRECOMPILATION_ARGS"], "")
        self.testcase_precompilation_args = ArgList(cfg["TESTCASE_PRECOMPILATION_ARGS"], "")
        self.testcase_compilation_args = ArgList(cfg["TESTCASE_COMPILATION_ARGS"], "")
        self.testcase_runtime_args = ArgList(cfg["TESTCASE_RUNTIME_ARGS"], "")


def _read_config(config: Path, fallback_config: Optional[Path] = None) -> Mapping[str, Any]:
    if fallback_config is None:
        fallback_doc = None
    else:
        fallback_doc = parse(fallback_config.read_text())
    doc = parse(config.read_text() if config.exists() else "")
    if fallback_doc is not None:
        _config_union(doc, fallback_doc)
    return doc


def _config_union(cfg1: Container, cfg2: Container) -> None:
    """Only works for configs similar to default_config.toml"""
    for k, v in cfg2.items():
        if k not in cfg1:
            cfg1[k] = v
        else:
            for inner_k, inner_v in v.items():
                if inner_k not in cfg1[k]:
                    cfg1[k][inner_k] = inner_v  # type: ignore



import asyncio
import re
import shutil
import sys
from pathlib import Path
from typing import Any, List, Mapping
from autograder.config_manager import GradingConfig

from autograder.testcase_utils.abstract_testcase import TestCase as AbstractTestCase
from autograder.testcase_utils.shell import ShellError, get_shell_command
from autograder.testcase_utils.submission import find_appropriate_source_file_stem
from autograder.util import AutograderError

PUBLIC_CLASS_MATCHER = re.compile(r"public(?:\w|\s)+class(?:\w|\s)+({)")

# Previously, we used SecurityManager to prevent reflection but now
# it's deprecated and we're forced to resort to this hack. It's not
# perfect but java does not leave us any other options.
REFLECTION_MATCHER = re.compile(r"import(\s+)java\.lang\.reflect\.")

# Java does not allow us to properly unset environment variables
# which is why we must use JNA to do so. However, because we are
# forced to insert our TestHelper into TestCase class, we cannot
# define any imports on top of it. Hence we have to use this hack.
LIBRARIES_REQUIRED_FOR_UNSETENV = """
import com.sun.jna.Library;
import com.sun.jna.Native;
import java.lang.reflect.Field;
import java.util.Map;
import java.util.HashMap;

"""
if sys.platform.startswith("win32"):
    ADDITIONAL_TEST_HELPER_KWARGS = {"SETENV": "_putenv", "C_LIBRARY": "msvcrt"}
    CLASSPATH = ".;jna.jar"
else:
    ADDITIONAL_TEST_HELPER_KWARGS = {"SETENV": "setenv", "C_LIBRARY": "c"}
    CLASSPATH = ".:*"


EXTRA_DIR = Path(__file__).parent / "extra"
PATH_TO_JNA_FILE = EXTRA_DIR / "jna.jar"


class TestCase(AbstractTestCase):
    """Please, ask students to remove their main as it can theoretically
    generate errors (not sure how though).
    Java doesn't support testcase precompilation because it must always
    link files on compilation.
    """

    source_suffix = ".java"
    executable_suffix = ""
    helper_module = "TestHelper.java"  # type: ignore
    compiler = get_shell_command("javac")
    virtual_machine = get_shell_command("java")

    @classmethod
    def is_installed(cls) -> bool:
        return cls.compiler is not None and cls.virtual_machine is not None

    @classmethod
    async def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        cli_args: str,
        config: Mapping[str, Any],
        lock: asyncio.Lock,
    ):
        stem = find_appropriate_source_file_stem(submission, possible_source_file_stems)
        if stem is None:
            raise AutograderError(
                f"Submission {submission} has an inappropriate file name. Please, specify POSSIBLE_SOURCE_FILE_STEMS in config.ini"
            )
        copied_submission = await super().precompile_submission(submission, student_dir, [stem], cli_args, config, lock)
        try:
            if not REFLECTION_MATCHER.search(copied_submission.read_text()):
                await cls.compiler(copied_submission, *cli_args.split())
            else:
                raise ShellError(1, "The use of reflection is forbidden in student submissions.")
        finally:
            copied_submission.unlink()

        return copied_submission.with_suffix(".class")

    async def compile_testcase(self, precompiled_submission: Path, cli_args: str):
        new_self_path = precompiled_submission.with_name(self.path.name)
        await self.compiler(
            new_self_path,
            "-cp",
            CLASSPATH,
            *cli_args.split(),
            cwd=precompiled_submission.parent,
        )
        return lambda *args, **kwargs: self.virtual_machine("-cp", CLASSPATH, self.path.stem, *args, **kwargs)

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        shutil.copyfile(PATH_TO_JNA_FILE, student_dir / PATH_TO_JNA_FILE.name)

    def delete_executable_files(self, precompiled_submission: Path):
        for p in precompiled_submission.parent.iterdir():
            if p.suffix == ".class" and self.path.stem in p.stem:
                p.unlink()

    def prepend_test_helper(self):
        """Puts private TestHelper at the end of testcase class.
        This is quite a crude way to do it but it is the easiest
        I found so far.
        """
        content = self.path.read_text()
        formatted_test_helper = self.get_formatted_test_helper(**ADDITIONAL_TEST_HELPER_KWARGS)
        final_content = self._add_at_the_beginning_of_public_class(formatted_test_helper, content)
        self.path.write_text(LIBRARIES_REQUIRED_FOR_UNSETENV + final_content)

    def _add_at_the_beginning_of_public_class(self, helper_class: str, java_file: str):
        """Looks for the first bracket of the first public class and inserts test helper next to it"""
        # This way is rather crude and can be prone to errors if left untested,
        # but java does not really leave us any other way to do it.
        match = PUBLIC_CLASS_MATCHER.search(java_file)
        if not match:
            raise ValueError(f"Public class not found in {self.path}")
        else:
            main_class_index = match.end(1)

        return "".join(
            [
                java_file[:main_class_index],
                "\n" + helper_class + "\n",
                java_file[main_class_index:],
            ]
        )



# Seems like too much work for only 2 simple functions, no?


import random
import string
from typing import Tuple

VALIDATING_STRING_LENGTH = 30
# DO NOT ATTEMPT TO ADD string.punctuation: it might break strings in some languages because it contains "" and ''
POSSIBLE_CHARS_IN_VALIDATING_STRING = string.ascii_letters + string.digits
LAST_LINE_SPLITTING_CHARACTER = "$"


def generate_validating_string() -> str:
    """ Generates a string of pseudo-random ascii characters """
    # As far as I understand, we do not need this string to be cryptographically strong
    return "".join(random.choice(POSSIBLE_CHARS_IN_VALIDATING_STRING) for _ in range(VALIDATING_STRING_LENGTH))


# Beware: Horrible space and time complexity. Maybe we should optimize that?
def validate_output(output: str, required_validating_string: str) -> Tuple[str, float, bool]:
    """Checks if the output contains the correct validating string on the last line

    Returns the output without the last line (or empty string if it was empty),
    the score that was printed with the validation string and boolean that is true if output is valid.
    """
    if not output:
        return output, 0, False
    split_output = output.splitlines()
    last_line = split_output.pop(-1)
    try:
        result, actual_validating_string = last_line.split(LAST_LINE_SPLITTING_CHARACTER, 1)
        result = float(result)
    except ValueError:  # Means that the last line is not formatted correctly
        return output, 0, False
    output = "\n".join(split_output)
    return output, result, actual_validating_string == required_validating_string



# These functions will format both the expected and student's outputs to compare them against each other
# It is not required to specify any outputs for your testcase_utils. I.e. If you want, you do not even have to create
# stdout_formatters.py in your directory or you can leave it blank. Type hints are also not required and are only
# here for reference.

# For specific (not ALL) formatters, you name them after their testcase files. I.e. If testcase file has a name
# "TestFactorial.java", the formatter's name will be TestFactorial(s: str)
# (function argument name can be anything and is unimportant).


# def DEFAULT(s: str) -> str:
#     """This formatter will be used on all testcase outputs
#     by default if no other formatters are specified for them

#     This formatter will only leave digits in the output because the author
#     of it decided that any other text is unimportant for the purposes of his testing.
#     """
#     return "".join(filter(lambda c: c.isdigit(), s))


# def testcase1(s: str) -> str:
#     """Only this formatter will be applied to testcase1
#     It will remove all digits from output.
#     """
#     return "".join(filter(lambda c: not c.isdigit(), s))


# def testcase2(s: str) -> str:
#     """Only this formatter will be applied to testcase2
#     It will remove all spaces from output.
#     """
#     return s.replace(" ", "")


# def testcase3(s: str) -> str:
#     """Only this formatter will be applied to testcase3
#     It will remove all whitespace from output.
#     """
#     return "".join(filter(lambda c: not c.isspace(), s))



import asyncio
import py_compile
import sys
from pathlib import Path
from typing import Any, List, Mapping
from autograder.config_manager import GradingConfig

from autograder.testcase_utils.abstract_testcase import TestCase as AbstractTestCase
from autograder.testcase_utils.shell import EMPTY_COMMAND, get_shell_command


class TestCase(AbstractTestCase):
    """A proof of concept of how easy it is to add new languages.
    Will only work if python is accessible via python3 alias for now.
    """

    source_suffix = ".py"
    executable_suffix = ".pyc"
    helper_module = "test_helper.py"  # type: ignore
    interpreter = get_shell_command(sys.executable)

    @classmethod
    def is_installed(cls) -> bool:
        return cls.interpreter is not EMPTY_COMMAND

    @classmethod
    async def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        cli_args: str,
        config: Mapping[str, Any],
        lock: asyncio.Lock,
    ):
        copied_submission = await super().precompile_submission(
            submission,
            student_dir,
            [submission.stem],
            cli_args,
            config,
            lock,
        )
        kwargs = {}
        if "-O" in cli_args:
            kwargs["optimize"] = 1
        executable_path = copied_submission.with_suffix(cls.executable_suffix)
        py_compile.compile(file=str(copied_submission), cfile=str(executable_path), doraise=True)
        copied_submission.unlink()
        return executable_path

    async def compile_testcase(self, precompiled_submission: Path, cLi_args: str):
        return lambda *args, **kwargs: self.interpreter(
            self.make_executable_path(precompiled_submission),
            *args,
            env={"STUDENT_SUBMISSION": precompiled_submission.stem, **kwargs.pop("env")},
            **kwargs,
        )

    async def precompile_testcase(self, cli_args: str):
        kwargs = {}
        if "-O" in cli_args:
            kwargs["optimize"] = 1
        executable_path = self.path.with_suffix(self.executable_suffix)
        py_compile.compile(file=str(self.path), cfile=str(executable_path), doraise=True, **kwargs)
        self.path.unlink()
        self.path = executable_path

    def delete_source_file(self, source_path: Path):
        """Source file is the same as executable file so we don't need to delete it"""

    def make_executable_path(self, submission: Path) -> Path:
        return submission.with_name(self.path.name)



import asyncio
import dataclasses
import shutil
import sys
from abc import ABC, ABCMeta, abstractmethod
from concurrent.futures import TimeoutError
from inspect import getsourcefile
from pathlib import Path
from typing import Any, Dict, List, Mapping, Tuple

from ..config_manager import GradingConfig
from .exit_codes import SYSTEM_RESERVED_EXIT_CODES, USED_EXIT_CODES, ExitCodeEventType
from .shell import ShellCommand, ShellError
from .test_helper_formatter import get_formatted_test_helper
from .testcase_io import TestCaseIO
from .testcase_result_validator import generate_validating_string, validate_output


@dataclasses.dataclass
class TestCaseResult:
    grade: float
    message: str
    extra_output_fields: Dict[str, str] = dataclasses.field(default_factory=dict)


class SourceDirSaver(ABCMeta, type):
    """Useful in getting the resources associated with each testcase type"""

    type_source_file: Path

    def __new__(mcs, name, bases, dct):  # type: ignore
        cls = super().__new__(mcs, name, bases, dct)
        # We needed a way to get a source file based solely on __class__ to access its sibling directories
        source_file = getsourcefile(cls)
        if source_file is None:
            raise FileNotFoundError(f"Source file for class {cls} has not been found.")
        cls.type_source_file = Path(source_file)
        return cls


class TestCase(ABC, metaclass=SourceDirSaver):
    source_suffix = ".source_suffix"  # dummy value
    executable_suffix = ".executable_suffix"  # dummy value

    type_source_file: Path
    test_helpers_dir: Path
    path: Path
    weight: float
    max_score: int
    io: TestCaseIO
    validating_string: str

    config: Mapping[str, Any]

    # Note that this structure will not work for any children of this class until 3.9
    # because classmethod does not wrap property correctly until then.
    # See https://bugs.python.org/issue19072
    @classmethod
    @property
    @abstractmethod
    def helper_module(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def is_installed(cls) -> bool:
        """Returns True if software necessary to run the testcase is installed on the system"""

    @classmethod
    def get_template_dir(cls):
        return cls.type_source_file.parent / "templates"

    @abstractmethod
    async def compile_testcase(self, precompiled_submission: Path, cli_args: str) -> ShellCommand:
        """Compiles student submission and testcase into a single executable
        (or simply returns the command to run the testcase if no further compilation is necessary)

        pwd = temp/student_dir
        """

    def __init__(
        self,
        path: Path,
        timeout: float,
        weight: float,
        io: TestCaseIO,
        config: Mapping[str, Any],
    ):
        self.test_helpers_dir = self.type_source_file.parent / "helpers"
        self.path = path
        self.timeout = timeout
        self.weight = weight
        self.max_score = int(weight * 100)

        self.name = path.name

        self.io = io
        self.validating_string = generate_validating_string()

        self.config = config

        self.prepend_test_helper()

    @classmethod
    async def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        cli_args: str,
        config: Mapping[str, Any],
        # Lock is used to allow for optimizations during precompilation.
        # For example, when all submissions share a single library that
        # needs to be precompiled only once.
        lock: asyncio.Lock,
    ) -> Path:
        """Copies student submission into student_dir and either precompiles
        it and returns the path to the precompiled submission or to the
        copied submission if no precompilation is necessary

        pwd = temp/student_dir
        """
        destination = (student_dir / possible_source_file_stems[0]).with_suffix(cls.source_suffix)
        shutil.copy(str(submission), str(destination))
        return destination

    async def precompile_testcase(self, cli_args: str):
        """Replaces the original testcase file with its compiled version,
        thus making reading its contents as plaintext harder.
        Useful in preventing cheating.

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        """Do nothing by default"""
        pass

    @classmethod
    def is_a_type_of(cls, file: Path, possible_source_file_stems: List[str]) -> bool:
        return file.suffix == cls.source_suffix

    def get_path_to_helper_module(self) -> Path:
        return self.test_helpers_dir / self.helper_module

    async def run(
        self,
        precompiled_submission: Path,
        testcase_compilation_args: str,
        testcase_runtime_args: str,
    ) -> TestCaseResult:
        """Returns student score and message to be displayed"""
        shutil.copy(self.path, precompiled_submission.with_name(self.path.name))
        try:
            test_executable = await self.compile_testcase(precompiled_submission, testcase_compilation_args)
        except ShellError as e:
            return TestCaseResult(0, e.format("Failed to compile"))

        result = await self._weightless_run(precompiled_submission, test_executable, testcase_runtime_args)
        result.grade *= self.weight

        self.delete_executable_files(precompiled_submission)
        return result

    def make_executable_path(self, submission: Path) -> Path:
        """By combining test name and student name, it makes a unique path"""
        return submission.with_name(self.path.stem + submission.stem + self.executable_suffix)

    def prepend_test_helper(self):
        """Prepends all of the associated test_helper code to test code

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """
        with self.path.open() as f:
            content = f.read()
            final_content = self.get_formatted_test_helper() + "\n" + content
        with self.path.open("w") as f:
            f.write(final_content)

    def get_formatted_test_helper(self, **exta_format_kwargs: str) -> str:
        return get_formatted_test_helper(self.get_path_to_helper_module(), **exta_format_kwargs)

    def delete_executable_files(self, precompiled_submission: Path):
        path = self.make_executable_path(precompiled_submission)
        # Windows sucks at deleting such files
        if path.exists() and sys.platform != "win32":
            path.unlink()

    async def _weightless_run(
        self,
        precompiled_submission: Path,
        compiled_testcase: ShellCommand,
        testcase_runtime_args: str,
    ) -> TestCaseResult:
        """Returns student score (without applying testcase weight) and message to be displayed"""
        try:
            result = await compiled_testcase(
                *testcase_runtime_args.split(),
                stdin=self.io.input,
                timeout=self.timeout,
                cwd=precompiled_submission.parent,
                env={"VALIDATING_STRING": self.validating_string},
                allowed_exit_codes=USED_EXIT_CODES,
            )
            exit_code = result.returncode
        except TimeoutError:
            return TestCaseResult(0, f"Exceeded time limit of {self.timeout} seconds")
        except ShellError as e:
            return TestCaseResult(0, f"Crashed due to signal {e.returncode}:\n{e.stderr}\n")
        raw_output = result.stdout
        output, score, output_is_valid = validate_output(raw_output, self.validating_string)
        if not output_is_valid:
            # This  means that either the student used built-in exit function himself
            # or some testcase helper is broken, or a testcase exits itself without
            # the use of helper functions.
            return TestCaseResult(
                0,
                "None of the helper functions have been called.\n"
                f"Instead, exit() has been called with exit_code {exit_code}.\n"
                "It could indicate student cheating or testcase_utils being written incorrectly.",
            )
        elif exit_code == ExitCodeEventType.CHECK_STDOUT:
            if self.io.expected_output_equals(output):
                return TestCaseResult(100, f"{int(100 * self.weight)}/{self.max_score}")
            else:
                return TestCaseResult(0, f"0/{self.max_score} (Wrong output)")
        elif exit_code == ExitCodeEventType.RESULT:
            weighted_score = round(score * self.weight, 2)
            # We do this to make output prettier in case the student gets full points
            if weighted_score == self.max_score:
                weighted_score = round(weighted_score)
            message = f"{weighted_score}/{self.max_score}"
            if score == 0:
                message += " (Wrong answer)"
            return TestCaseResult(score, message)
        elif exit_code in SYSTEM_RESERVED_EXIT_CODES or exit_code < 0:
            # We should already handle this case in try, except block. Maybe we need more info in the error?
            raise NotImplementedError(f"System error with exit code {exit_code} has not been handled.")
        else:
            raise ValueError(f"Unknown system code {exit_code} has not been handled.")



import math
from pathlib import Path
from typing import List

import numpy as np
from antlr4 import *

from .comparison import get_similarity
from .lexers.CLexer import CLexer
from .lexers.CppLexer import CppLexer
from .lexers.Java8Lexer import JavaLexer
from .lexers.Python3Lexer import Python3Lexer


# entry point function that is called to compare a set of files with each other
def compare(paths: List[Path]) -> dict:
    numFiles = len(paths)
    if numFiles == 0:
        raise ValueError("No files found")
    language_map = initialize_language(paths)
    results = {}
    for language in language_map:
        files = language_map[language]["files"]
        if len(files) == 0:
            continue
        # get token stream for each file and total frequency for each token type
        parsed_file_map = parse_files(language_map[language])
        token_streams = parsed_file_map["token_streams"]
        lengths = parsed_file_map["lengths"]
        # construct similarity matrix to weight the significance of matching tokens. Matching uncommon tokens
        # is weighted heavier as it is more likely to be a result of plagiarism
        similarityMatrix = build_similarity_matrix(parsed_file_map["freq"])
        # find the similarity score of comparing a file to itself. This is used to normalize the similarity score
        # calculated when comparing unique files
        self_similarities = build_self_similarities(token_streams, similarityMatrix, lengths)
        result = run_comparisons(token_streams, similarityMatrix, self_similarities, lengths)
        results[language] = convert_results(result, files)

    return results


# determine language of files and initialize language-specific variables
def initialize_language(paths: List[Path]) -> dict:
    language_partition = {"java": [], "py": [], "c": [], "cpp": []}

    for path in paths:
        language_partition[path.suffix.split(".")[-1]].append(path)

    # format: "file ending": [Lexer class used to parse, [ignore list of token types to ignore when parsing],
    # number of unique tokens in language]
    language_mapper = {
        "java": {
            "lexer": JavaLexer,
            "ignore_list": np.array([-1, 106, 107]),
            "num_tokens": 108,
        },
        "py": {
            "lexer": Python3Lexer,
            "ignore_list": np.array([-1, 39]),
            "num_tokens": 99,
        },
        "c": {
            "lexer": CLexer,
            "ignore_list": np.array([-1, 117, 118]),
            "num_tokens": 119,
        },
        "cpp": {
            "lexer": CppLexer,
            "ignore_list": np.array([-1, 144, 145]),
            "num_tokens": 146,
        },
    }
    # add files of each programming language to mapper object
    for key in language_mapper:
        language_mapper[key]["files"] = language_partition[key]
    return language_mapper


def parse_files(language: dict) -> dict:
    files = language["files"]
    lexer_class = language["lexer"]
    ignore_list = language["ignore_list"]
    num_tokens = language["num_tokens"]
    tokenStreams = []
    freq = [0 for x in range(num_tokens + 1)]
    total_tokens = 0
    for file in files:
        with file.open() as f:
            stream = InputStream(f.read())
            lexer = lexer_class(stream)
            tokens = CommonTokenStream(lexer)
            # TODO: look at fetch
            tokens.fetch(5000)
            array = []
            for token in tokens.tokens:
                # remove comments, blank lines, etc
                if token.type > num_tokens:
                    pass
                if token.type not in ignore_list and token.type <= num_tokens:
                    array += [token.type]
                    freq[token.type] += 1
                    total_tokens += 1
            tokenStreams += [np.array(array)]
    # find frequency of each token as percentage of total tokens
    if total_tokens == 0:
        freq = [1e-10 for x in range(len(freq))]
    else:
        for i in range(len(freq)):
            freq[i] = max(freq[i] / total_tokens, 1e-10)
    lengths = np.array([len(x) for x in tokenStreams])
    converted_token_stream = np.zeros((len(tokenStreams), np.max(lengths)), np.int32)
    for i in range(len(tokenStreams)):
        converted_token_stream[i, : lengths[i]] = tokenStreams[i]
    return {
        "token_streams": converted_token_stream,
        "freq": np.array(freq),
        "lengths": lengths,
    }


def build_similarity_matrix(freq: np.ndarray) -> np.ndarray:
    # alpha is how heavily matching tokens should be penalized, beta is how heavily
    # mismatching tokens should be rewarded. alpha + beta = 1
    alpha = 0.65
    beta = 0.35
    matrixLen = len(freq) + 1
    matrix = np.zeros((matrixLen, matrixLen))
    for i in range(matrixLen):
        for j in range(i, matrixLen):
            # matching a gap with a gap is impossible
            if i == matrixLen - 1 and j == matrixLen - 1:
                value = 1e10
            # value of matching token i with a gap
            elif j == matrixLen - 1:
                value = 4 * beta * math.log2(freq[i])
            # value of matching token i
            elif i == j:
                value = -1 * alpha * math.log2(freq[i] * freq[i])
            # value of matching two different tokens, i and j
            else:
                value = beta * math.log2(freq[i] * freq[j])
            # value of matching i and j is same as matching j and i
            matrix[i, j] = int(1000 * value)
            matrix[j, i] = int(1000 * value)
    return np.array(matrix)


def build_self_similarities(token_streams: np.ndarray, matrix: np.ndarray, lengths: np.ndarray) -> np.ndarray:
    self_similarities = []
    for i in range(len(token_streams)):
        score = 0
        # for each token, find value of matching that token with itself and add to sum
        for tok in token_streams[i, : lengths[i]]:
            score += matrix[int(tok), int(tok)]
        self_similarities.append(score)
    return np.array(self_similarities, dtype=np.int32)


def run_comparisons(
    token_streams: np.ndarray,
    similarity_matrix: np.ndarray,
    self_similarities: np.ndarray,
    lengths: np.ndarray,
) -> np.ndarray:
    similarity_scores = np.zeros((len(token_streams), len(token_streams)))
    for i in range(len(token_streams)):
        for j in range(i + 1, len(token_streams)):
            similarity = get_similarity(
                token_streams[i, : lengths[i]],
                token_streams[j, : lengths[j]],
                similarity_matrix,
                self_similarities[i] + self_similarities[j],
            )
            similarity_scores[i, j] = similarity
    return similarity_scores


def convert_results(results: np.ndarray, files: List[Path]) -> dict:
    converted_results = {}
    for i in range(len(results)):
        for j in range(i + 1, len(results)):
            converted_results[frozenset((files[i], files[j]))] = results[i, j]
    return converted_results




# Allows us to use student's module from our testcase
import importlib
import os
import sys


def CHECK_STDOUT():
    print(f"\n-1{ % SPLITCHAR % }{VALIDATING_STRING}", end="")
    exit({ % CHECK_STDOUT_EXIT_CODE % })

def RESULT(r: float):
    print(f"\n{r}{ % SPLITCHAR % }{VALIDATING_STRING}", end="")
    exit({ % RESULT_EXIT_CODE % })

def PASS():
    RESULT(100)

def FAIL():
    RESULT(0)

def CHEATING_ATTEMPT():
    exit({ % CHEAT_ATTEMPT_EXIT_CODE % })


VALIDATING_STRING = os.environ.pop("VALIDATING_STRING", None)
if __name__ != "__main__" or VALIDATING_STRING is None:
    # If a student tries to import a testcase, the test will instantly fail.
    # It would also technically be a recursive import which would most likely
    # raise an exception but I put this here as a safeguard.
    print("CHEATING ATTEMPT")
    del VALIDATING_STRING
    CHEATING_ATTEMPT()

student_submission = importlib.import_module(os.environ.pop("STUDENT_SUBMISSION", None))


__title__ = "assignment-autograder"
__description__ = "Automatic assignment grading for instructor use in programming courses"
__version__ = "3.3.1"
__author__ = "Stanislav Zmiev"
__author_email__ = "szmiev2000@gmail.com"
__license__ = "GPL-3.0"



import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional


L = logging.getLogger("AUTOGRADER")
logging.basicConfig(level=logging.CRITICAL, handlers=[logging.StreamHandler()])


def main(argv: Optional[List[str]] = None):
    """Returns the average score of the students"""
    if argv is None:
        argv = sys.argv[1:]

    parser = _create_parser()

    args = parser.parse_args(argv)
    if args.version:
        from autograder.__version__ import __version__

        print(__version__)
        exit(0)
    # the interface architecture needs to be refactored a bit. For now, this hack with getattr
    # will prevent errors if autograder has been called on its own.
    elif getattr(args, "submission_path", None) is None:
        parser.print_help()
        exit()

    current_dir = (Path.cwd() / args.submission_path).resolve()
    return _evaluate_args(args, current_dir)


def _create_parser():
    parser = argparse.ArgumentParser(prog="autograder")
    parser.add_argument("-V", "--version", action="store_true", help="print the autograder version number and exit")
    subparsers = parser.add_subparsers(title="Commands", dest="command")
    _create_run_parser(subparsers)
    _create_stats_parser(subparsers)
    _create_guide_parser(subparsers)
    _create_plagiarism_parser(subparsers)
    return parser


def _create_run_parser(subparsers):
    parser = subparsers.add_parser("run", help="Grade submissions in submission path or in current directory")
    parser.add_argument("-j", "--json", action="store_true", help="Output grades in json format")
    parser.add_argument("-v", "--verbose", action="store_true", help="Show all debugging output")
    _add_submission_path_argument(parser)
    _add_submission_list_argument(parser)


def _create_stats_parser(subparsers):
    # TODO: Rewrite stats parser to handle json output
    # parser = subparsers.add_parser("stats", help="Display statistics on student grades")
    # parser.add_argument(
    #     "-p",
    #     "--print",
    #     type=float,
    #     nargs="?",
    #     default=None,
    #     const=100,
    #     metavar="min_score",
    #     help="Use after already graded to print assignments with score >= min_score",
    # )
    # _add_submission_path_argument(parser)
    pass


def _create_guide_parser(subparsers):
    parser = subparsers.add_parser("guide", help="Guide you through setting up a grading environment")
    _add_submission_path_argument(parser)


def _create_plagiarism_parser(subparsers):
    parser = subparsers.add_parser("plagiarism", help="Checks how similar the submissions are to each other")
    _add_submission_path_argument(parser)
    _add_submission_list_argument(parser)


def _add_submission_path_argument(parser: argparse.ArgumentParser):
    parser.add_argument(
        "submission_path",
        type=Path,
        nargs="?",
        default=Path.cwd(),
        help="Path to directory that contains student submissions",
    )


def _add_submission_list_argument(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-s",
        "--submissions",
        action="store",
        nargs="*",
        metavar="<name>",
        type=Path,
        default=[],
        help="Only consider submissions with specified file names (without full path)",
    )


def _evaluate_args(args: argparse.Namespace, current_dir: Path):
    if sys.platform.startswith("darwin") and not args.json:
        print("OSX is not officially supported. Proceed with caution.")
    from autograder.autograder import AutograderPaths, Grader

    if args.command == "guide":
        from autograder import guide

        guide.main(AutograderPaths(current_dir))
    elif args.command == "run":
        if args.verbose:
            L.setLevel(logging.DEBUG)
        submissions = [s.name for s in args.submissions]
        return Grader(current_dir, args.json, submissions).run()
    elif args.command == "plagiarism":
        import json

        from . import plagiarism_detection

        files = [f for f in current_dir.iterdir() if f.is_file() and not f.suffix.endswith(".txt")]
        if args.submissions:
            submissions = [s.name for s in args.submissions]
            files = [f for f in files if f.name in submissions]
        result = plagiarism_detection.compare(files)
        result = {tuple(k): v for k, v in result[list(result.keys())[0]].items()}
        # print(result)
        output = [{"student1": k[0].name, "student2": k[1].name, "similarity_score": v} for k, v in result.items()]
        output.sort(key=lambda v: v["similarity_score"], reverse=True)
        print(json.dumps({"results": output}, indent=4))
    else:
        raise NotImplementedError(
            f"Unknown command '{args.command}' supplied.\nTry 'autograder --help for more information'"
        )
    return -1


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autograder',
 'autograder.plagiarism_detection',
 'autograder.plagiarism_detection.lexers',
 'autograder.testcase_types.cpython',
 'autograder.testcase_types.cpython.helpers',
 'autograder.testcase_types.cpython.templates',
 'autograder.testcase_types.gcc',
 'autograder.testcase_types.javac',
 'autograder.testcase_utils']

package_data = \
{'': ['*'],
 'autograder': ['testcase_types/gcc/c++_templates/*',
                'testcase_types/gcc/c_templates/*',
                'testcase_types/gcc/helpers/*',
                'testcase_types/gcc/memleak/*',
                'testcase_types/javac/extra/*',
                'testcase_types/javac/helpers/*',
                'testcase_types/javac/templates/*'],
 'autograder.plagiarism_detection': ['antlr_grammars/*']}

install_requires = \
['antlr4-python3-runtime==4.9.2',
 'numba>=0.54.1,<0.55.0',
 'tomlkit>=0.7.2,<0.8.0']

entry_points = \
{'console_scripts': ['autograder = autograder.__main__:main']}

setup_kwargs = {
    'name': 'assignment-autograder',
    'version': '3.3.1',
    'description': 'Automatic assignment grading for instructor use in programming courses',
    'long_description': 'This utility aims to provide a simple, yet secure and highly configurable way to autograde programming assignments.\n\nI consider it to be finished. Autograder has been tested on a real university class with hundreds of students and has shown to be error-less (in terms of grades), fast, and protected from cheating.\n#### Table of Contents  \n[Features](#Features) \n[Installation](#Platform-Support)  \n[Installation](#Installation)   \n[Supported Programming Languages](#Supported-Programming-Languages)   \n[Quickstart](#Quickstart)   \n[Usage](#Usage)   \n[Writing testcases](#Writing-testcases)  \n[Helper functions](#Helper-functions)  \n[Limitations](#Limitations)   \n[Anti Cheating](#Anti-Cheating)   \n[Adding Programming Languages](#Adding-Programming-Languages)\n# Features\n* Most features are demonstrated in examples/ directory\n* [Easy to grade](#Usage) \n* [Easy-to-write testcases](#Writing-testcases)  \n* Testcase grade can be based on [student\'s stdout](#Helper-functions)\n* Can grade C, C++, Java, and Python code\n* A file with testcase results will be generated for each student\n* You can customize the total points for the assignment, maximum running time of student\'s program, file names to be considered for grading, formatters for checking student stdout, and [much more](https://github.com/Ovsyanka83/autograder/blob/master/autograder/default_config.ini).\n* [Anti Cheating capabilities](#Anti-Cheating) that make it nearly impossible for students to break the grader and choose their grades.\n* Grading submissions in multiple programming languages at once, as long as there are testcases written in each language.\n* Most of these features are described in detail in [default_config.ini](https://github.com/Ovsyanka83/autograder/blob/master/autograder/default_config.ini) and demonstrated in examples/ directory.\n* Stdout-only (language-agnostic) grading supported\n* JSON result output supported if autograder needs to be integrated as a part of a larger utility\n# Platform Support\n* Linux is fully supported\n* OS X has not been tested\n* Windows is partially supported:\n  * C/C++ have only been tested with `mingw` installed with `chocolatey`\n  * Java is only supported if autograder is run with Administrative privileges\n  * Stdout-testcases that require compilation are only supported if `make` is installed\n  * Stdout-testcases that require shebang lines are not supported\n# Installation\n* Currently Python >= 3.7 is necessary.\n* Run `pip3 install assignment-autograder`\n* If you want to update to a newer version, run `pip3 install --upgrade --no-cache-dir assignment-autograder`\n# Supported Programming Languages\n* Java (only through javac and java alias)\n* C (only through gcc)\n* C++ (only through g++)\n* CPython (3.6-3.10)\n* Any programming language if stdout-only grading is used\n# Quickstart\n* Run `autograder guide path/to/directory/you\'d/like/to/grade`. The guide will create all of the necessary configurations and directories for grading and will explain how to grade.\n* Read [Usage](#Usage) section\n# Usage\n1) Create tests directory in the same directory as student submissions. Its structure is shown in [examples](https://github.com/Ovsyanka83/autograder/tree/master/examples). (can be automatically created using [--guide](#Quickstart))\n1) __Optional__ files that can be automatically created by [--guide](#Quickstart) CLI option and whose use is demostrated by [examples](https://github.com/Ovsyanka83/autograder/tree/master/examples):\n    1) Input (stdin) and expected output (__stdout__) text files in their respective directories for each testcase. If a test does not require input and/or stdout, the respective text file is also not required.\n    1) Create [config.ini](https://github.com/Ovsyanka83/autograder/blob/master/autograder/default_config.ini) and change configuration to fit your needs (If you do not include some fields, autograder will use the respective fields from default_config.ini)\n    1) Create [stdout_formatters.py](https://github.com/Ovsyanka83/autograder/blob/master/autograder/default_stdout_formatters.py) and edit it to fit your needs. They will format student\'s stdout to allow you to give credit to students even if their stdout is not exactly the same as expected.\n1) Write testcases as described [below](#Writing-testcases) using [examples](https://github.com/Ovsyanka83/autograder/tree/master/examples) as reference.\n1) Run `autograder run path/to/submissions/dir` from command line.\n## Writing testcases\n* Write a main that follows the same structure as one of the examples in your programming language. The main should usually call student\'s code, check its result, and call one of the helper functions (when working with stdout, you don\'t check the result, and simply allow autograder to handle grading by calling CHECK_STDOUT())\n* Assume that student\'s code is available in your namespace. Examples demonstrate exactly how to call students\' functions.\n* Assume that helper functions CHECK_STDOUT(), RESULT(int r), PASS(), FAIL() are predefined and use them to return student scores to the grader\n* Each helper function prints the student\'s score, __validation string__, terminates the execution of the program and returns its respective exit code that signifies to autograder if the testcase ended in a result, cheating attempt, or if stdout checking is necessary.\n* Each testcase is graded out of 100% and each grade is a 64bit double precision floating point number, which means that you can fully control how much partial credit is given in non-stdout checking tests.\n### Helper functions\n  * CHECK_STDOUT() indicates that we do not check student\'s return values for the testcase and that we only care about their output (__stdout__) that will be checked by the autograder automatically using student\'s stdout and the output files with the same name stem as the testcase. (beware: printing anything within your testcase will break this functionality)\n  * RESULT(double r) returns student\'s score r back to the grader (0 - 100)\n  * PASS() returns the score of 100% back to the grader and is equivalent to RESULT(100)\n  * FAIL() returns the score of 0% back to the grader and is equivalent to RESULT(0)\n## Limitations\n* At the point of writing this readme, stdout checking is a PASS or FAIL process (i.e. no partial credit possible). The reason is that allowing for \'partial similarity\' of outputs is too error-prone and could yield too many points for students that did not actually complete the task properly. If you want to increase the chances of students\' stdout matching, you should use stdout formatters described [above](#Usage).\n* If you don\'t prototype student functions you want to test in your C/C++ testcases, you will run into undefined behavior because of how C and C++ handle linking.\n* __Student\'s main functions ARE NOT meant to be accessed because testcase must be the starting point of the program.__ They are, however, accessible if necessary but undocumented in general case and always accessible in stdout-only grading.\n## Anti Cheating\nOne of the main weaknesses of automatic grading is how prone it is to cheating. Autograder tries to solve this problem with methods described in this section. Currently, (as far as I\'ve read and tested), it is impossible to cheat autograder. However, Java might still have some weird ways of doing this but there are protections against all of the most popular scenarios (decompiling and parsing testcases, using System.exit, trying to read security key from environment variables, using reflection to use private members of the test helper)\n* To restrict the student from exiting the process himself and printing the grade of his/her choice, I validate testcase stdout using a pseudorandom key called __validation string__. Autograder gives the string to the testcase as an environment variable which is erased right after the testcase saves it, and then it is automatically printed on the last line of stdout before the testcase exits. The autograder, then, pops it from stdout and verifies that it is the same string it sent. If it is not, the student will get the respective error message and a 0 on the testcase.\n* To prevent students from simply importing the string from the testcase file, test helper files (described above) all have some way of disallowing imports. For C/C++, it is the static identifier, for Java, it is the private method modifiers and SecurityManager to protect against reflection, for python it is throwing an error if __name__ != "__main__". I assume that similar precautions can be implemented in almost any language added into autograder.\n* Simply parsing validating string from the testcase file is impossible because it is passed at runtime.\n* As an additional (and maybe unnecessary) security measure, autograder precompiles testcases without linking for all languages except for java, thus decreasing the possibility that the student will simply parse the testcase file and figure out the correct return values if the security measure above doesn\'t work.\n\n# Adding Programming Languages\n* If you want to add a new language for grading, you have to:\n  1. Create a new directory in autograder/testcase_types/\n  2. Create a python module in that directory that contains a subclass of TestCase (from autograder/testcase_utils/abstract_testcase.py)\n  3. Create a helpers directory and write your test helper\n  4. Optionally, add the extra (extra files to be available to each testcase) and templates (examples of testcases written using the new language) directories\n* Use the other testcase subclasses and test helpers as reference\n* This point is optional but if you want full anti-cheating capabilities for your new language, you will need to consider three things:\n              \n  * Does your language support getting and unsetting environment variables? It is required to save validating string in your code without leaking it to students.\n  * Does your language support private-to-file functions/classes/methods/variables? It is required to prevent the student from simply importing helper functions and validating string.\n  * Does your language support precompilation (conversion to bytecode without linking)? It is not as important as other points but could speed up grading and hide testcase code from students. ',
    'author': 'Ovsyanka',
    'author_email': 'szmiev2000@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Ovsyanka83/autograder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



# Controls output to stdout and to output file


import json
from collections import deque
from pathlib import Path
from typing import Any, Deque, Sequence

from .testcase_utils.submission import Submission

KEY = """
\nKey:
\tFailed to Compile: Your submission did not compile due to a syntax or naming error
\tCompiled with warnings: Your submission uses unchecked or unsafe operations
\tCrashed due to signal SIGNAL_CODE: Your submission threw an uncaught exception.
\tAll signal error codes are described here: https://man7.org/linux/man-pages/man7/signal.7.html
\tExceeded Time Limit: Your submission took too much time to run (probably an infinite loop)
"""


def _empty_func(*a: Any, **kw: Any):
    """Python can't pickle lambdas so we use this
    Edit: Now pickling is not necessary but who cares anymore?
    """
    pass


# TODO: Replace path_to_results_dir and generate_results with path_to_results_dir: Optional[Path]
class GradingOutputLogger:
    LOGGER_NAME = "Autograder"

    def __init__(
        self,
        path_to_results_dir: Path,
        assignment_name: str,
        total_points_possible: int,
        generate_results: bool,
    ):
        self.results_dir = path_to_results_dir
        self.assignment_name = assignment_name
        self.total_points_possible = total_points_possible

        if not generate_results:
            self._silence_generating_results()

    def print_single_student_grading_results(self, submission: Submission):
        student_final_result = f"{submission.final_grade}/{self.total_points_possible}"
        self._print_single_student_grading_results_to_stdout(submission, student_final_result)
        self._print_single_student_grading_results_to_file(submission, student_final_result)

    def print_final_score(self, submissions: Sequence[Submission], score: int):
        print(f"\nAverage score: {score}/{self.total_points_possible}")

    def print_key(self):
        print(KEY)

    def _print_single_student_grading_results_to_stdout(self, submission: Submission, formatted_student_score: str):
        print(f"Grading {submission.name}")
        if submission.precompilation_error:
            print(f"{submission.precompilation_error}")
        else:
            for test_name, grade in submission.grades.items():
                additional_output = "\n".join(grade.extra_output_fields.values())
                if additional_output:
                    additional_output = "\n" + additional_output
                print(f"{test_name}: {grade.message}{additional_output}")
        print(f"\nResult: {formatted_student_score}\n\n")

    def _print_single_student_grading_results_to_file(self, submission: Submission, formatted_student_score: str):
        (self.results_dir / submission.old_path.name).write_text(
            self._format_output_for_student_file(submission, formatted_student_score)
        )

    def _silence_generating_results(self):
        self._print_single_student_grading_results_to_stdout = _empty_func  # type: ignore

    def _format_output_for_student_file(self, submission: Submission, formatted_student_score: str):
        """Replace this function with anything else if you want the output to have a different style"""
        str_builder: Deque[str] = deque()
        b = str_builder.append
        b(f"{self.assignment_name} Test Results\n\n")
        if not submission.precompilation_error:
            b("%-40s%s\n" % ("TestCase", "Result"))
        b("================================================================")
        if submission.precompilation_error:
            b("\n")
            b(submission.precompilation_error)
        else:
            for test_name, grade in submission.grades.items():
                additional_output = "\n".join(grade.extra_output_fields.values())
                b("\n%-40s%s" % (test_name, grade.message))
                if additional_output:
                    b("\n" + additional_output)
        b("\n================================================================\n")
        b("Result: " + formatted_student_score)
        b(KEY)
        return "".join(str_builder)


class JsonGradingOutputLogger(GradingOutputLogger):
    def __init__(
        self,
        path_to_results_dir: Path,
        assignment_name: str,
        total_points_possible: int,
        generate_results: bool,
    ):
        super().__init__(path_to_results_dir, assignment_name, total_points_possible, generate_results)

        self._print_single_student_grading_results_to_stdout = _empty_func

    def print_key(self):
        pass

    def print_final_score(self, submissions: Sequence[Submission], score: int):

        submission_results = [
            {
                "submission": s.old_path.name,
                "final_grade": s.final_grade,
                "testcase_scores": [
                    {
                        "name": test_name,
                        "message": grade.message,
                        **grade.extra_output_fields,
                    }
                    for test_name, grade in s.grades.items()
                ],
                "precompilation_error": s.precompilation_error,
            }
            for s in submissions
        ]
        submission_results.sort(key=lambda v: v["final_grade"], reverse=True)
        output = {
            "average_score": score,
            "total_points_possible": self.total_points_possible,
            "submissions": submission_results,
        }
        print(json.dumps(output, indent=4))



# student_submission is provided by autograder, represents student's module, and can be used freely
# PASS(), RESULT(res), FAIL(), and CHECK_STDOUT() are also provided by autograder and can be used freely


def main():
    # You can call any function from student's file like this.
    # It can have any arguments and return values -- use it like any other function.
    student_submission.some_student_function()

    # If you call this function, the program exits, and then autograder will compare this
    # program's stdout with whatever .txt file (with the same name stem as this test case)
    # you have provided in tests/output directory.
    CHECK_STDOUT()


main()



import numpy as np
from numba import njit


@njit
def get_similarity(a: np.ndarray, b: np.ndarray, matrix: np.ndarray, self_similarity_sum: int) -> float:
    dp = np.zeros((len(a), len(b)), dtype=np.int32)
    # initialize dp with score of matching first token of each token stream
    dp[0, 0] = max(0, matrix[a[0], b[0]])
    for i in range(1, len(a)):
        dp[i, 0] = max(0, dp[i - 1, 0] + matrix[a[i], -1])
    for i in range(1, len(b)):
        dp[0, i] = max(0, dp[0, i - 1] + matrix[-1, b[i]])
    for i in range(1, len(a)):
        for j in range(1, len(b)):
            # # score from matching both tokens
            maxScore = max(
                (
                    0,
                    dp[i - 1, j - 1] + matrix[a[i], b[j]],
                    dp[i - 1, j] + matrix[a[i], -1],
                    dp[i, j - 1] + matrix[-1, b[j]],
                )
            )
            dp[i, j] = maxScore
    return dp[-1, -1] * 2 / self_similarity_sum



# student_submission is provided by autograder, represents student's module, and can be used freely
# PASS(), RESULT(res), FAIL(), and CHECK_STDOUT() are also provided by autograder and can be used freely


def main():
    # You can call any function from student's file like this.
    # It can have any arguments and return values -- use it like any other function.
    result = student_submission.some_student_function()

    SOME_EXPECTED_RESULT = 83
    if result == SOME_EXPECTED_RESULT:
        PASS()
    else:
        FAIL()


main()



import importlib.util
import sys
from pathlib import Path
from types import ModuleType
from typing import Iterable


class AutograderError(Exception):
    """The base class for all exceptions that are raised from the incorrect use of Autograder"""


def import_from_path(module_name: str, path: Path) -> ModuleType:
    spec = importlib.util.spec_from_file_location(module_name, path)
    if not (spec and spec.loader):
        raise TypeError(f"File loader for {path} was not found. Please, refer to importlib docs.")
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    # exec_module is not always available which is why a linter won't be able to find it.
    # Docs do not state specific conditions for when it is available so if it's not, we
    # are okay with raising an ImportError.
    spec.loader.exec_module(module)
    return module


def get_file_names(dir_: Path) -> Iterable[str]:
    return (p.name for p in dir_.iterdir()) if dir_.exists() else ()


def hide_path_to_directory(string_to_hide_path_from: str, path_to_hide: Path, replacement: str = "...") -> str:
    """To prevent students from knowing the structure of instructor's directories, we remove the middle
    part of the path to student's file from output.

    Sometimes compilers output a longer/shorter path. For example, gcc through wsl will not start paths
    with "mnt/" while python will so we cut out the beginning and end of the path to hide the more important
    middle part of it.

    The length check is done to prevent indexing exceptions or corruption of output
    in case the path to remove is too short. For example,
    >>> "Hello. World.".replace(str(Path(".")), "...") == "Hello... World..."
    >>> "abc".replace("", "...") == ...a...b...c...
    """
    formatted_string = string_to_hide_path_from.replace(str(path_to_hide), replacement)
    if len(path_to_hide.parts) > 2:
        shorter_path_to_hide = Path("/".join(path_to_hide.parts[2:]))
        formatted_string = formatted_string.replace(str(shorter_path_to_hide), replacement)
    return formatted_string
